@@ -1,5 +1,3 @@
-from __future__ import print_function
-from __future__ import absolute_import
 import pycuda.autoinit
 import pycuda.gpuarray as gpuarray
 import pycuda.driver as cuda
