@@ -27,6 +27,7 @@ OTHER DEALINGS IN THE SOFTWARE.
 
 
 from pycuda.tools import context_dependent_memoize
+from typing import Any
 import numpy as np
 from pycuda.tools import dtype_to_ctype, VectorArg, ScalarArg
 from pytools import memoize_method
@@ -43,7 +44,6 @@ def get_elwise_module(
     after_loop="",
 ):
     from pycuda.compiler import SourceModule
-
     return SourceModule(
         """
         #include <pycuda-complex.hpp>
@@ -54,10 +54,10 @@ def get_elwise_module(
         __global__ void %(name)s(%(arguments)s)
         {
 
-          unsigned tid = threadIdx.x;
-          unsigned total_threads = gridDim.x*blockDim.x;
-          unsigned cta_start = blockDim.x*blockIdx.x;
-          unsigned i;
+          size_t tid = threadIdx.x;
+          size_t total_threads = gridDim.x*blockDim.x;
+          size_t cta_start = blockDim.x*blockIdx.x;
+          size_t i;
 
           %(loop_prep)s;
 
@@ -104,9 +104,9 @@ def get_elwise_range_module(
         extern "C"
         __global__ void %(name)s(%(arguments)s)
         {
-          unsigned tid = threadIdx.x;
-          unsigned total_threads = gridDim.x*blockDim.x;
-          unsigned cta_start = blockDim.x*blockIdx.x;
+          size_t tid = threadIdx.x;
+          size_t total_threads = gridDim.x*blockDim.x;
+          size_t cta_start = blockDim.x*blockIdx.x;
           long i;
 
           %(loop_prep)s;
@@ -463,8 +463,45 @@ def get_linear_combination_kernel(summand_descriptors, dtype_z):
     return func, tex_src
 
 
+def _get_real_dtype(dtype: "np.dtype[Any]") -> "np.dtype[Any]":
+    assert dtype.kind == "c"
+    return np.empty(0, dtype).real.dtype
+
+
 @context_dependent_memoize
-def get_axpbyz_kernel(dtype_x, dtype_y, dtype_z):
+def get_axpbyz_kernel(dtype_x, dtype_y, dtype_z,
+                    x_is_scalar=False, y_is_scalar=False):
+    """
+    Returns a kernel corresponding to ``z = ax + by``.
+
+    :arg x_is_scalar: A :class:`bool` which is *True* only if `x` is a scalar :class:`gpuarray`.
+    :arg y_is_scalar: A :class:`bool` which is *True* only if `y` is a scalar :class:`gpuarray`.
+    """
+    out_t = dtype_to_ctype(dtype_z)
+
+    # {{{ cast real scalars in context of complex scalar arithmetic
+
+    if dtype_z.kind == "c" and dtype_x.kind != "c":
+        dtype_z_real = _get_real_dtype(dtype_z)
+        x_t = dtype_to_ctype(dtype_z_real)
+    else:
+        x_t = out_t
+
+    if dtype_z.kind == "c" and dtype_y.kind != "c":
+        dtype_z_real = _get_real_dtype(dtype_z)
+        y_t = dtype_to_ctype(dtype_z_real)
+    else:
+        y_t = out_t
+
+    # }}}
+
+    x = "x[0]" if x_is_scalar else "x[i]"
+    a = f"({x_t}) a"
+    ax = f"{a}*(({x_t}) {x})"
+    y = "y[0]" if y_is_scalar else "y[i]"
+    b = f"({y_t}) b"
+    by = f"({b})*(({y_t}) {y})"
+    result = f"{ax} + {by}"
     return get_elwise_kernel(
         "%(tp_x)s a, %(tp_x)s *x, %(tp_y)s b, %(tp_y)s *y, %(tp_z)s *z"
         % {
@@ -472,7 +509,7 @@ def get_axpbyz_kernel(dtype_x, dtype_y, dtype_z):
             "tp_y": dtype_to_ctype(dtype_y),
             "tp_z": dtype_to_ctype(dtype_z),
         },
-        "z[i] = a*x[i] + b*y[i]",
+        f"z[i] = {result}",
         "axpbyz",
     )
 
@@ -488,7 +525,38 @@ def get_axpbz_kernel(dtype_x, dtype_z):
 
 
 @context_dependent_memoize
-def get_binary_op_kernel(dtype_x, dtype_y, dtype_z, operator):
+def get_binary_op_kernel(dtype_x, dtype_y, dtype_z, operator,
+                        x_is_scalar=False, y_is_scalar=False):
+    """
+    Returns a kernel corresponding to ``z = x (operator) y``.
+
+    :arg x_is_scalar: A :class:`bool` which is *True* only if `x` is a scalar :class:`gpuarray`.
+    :arg y_is_scalar: A :class:`bool` which is *True* only if `y` is a scalar :class:`gpuarray`.
+    """
+
+    out_t = dtype_to_ctype(dtype_z)
+
+    # {{{ cast real scalars in context of complex scalar arithmetic
+
+    if dtype_z.kind == "c" and dtype_x.kind != "c":
+        dtype_z_real = _get_real_dtype(dtype_z)
+        x_t = dtype_to_ctype(dtype_z_real)
+    else:
+        x_t = out_t
+
+    if dtype_z.kind == "c" and dtype_y.kind != "c":
+        dtype_z_real = _get_real_dtype(dtype_z)
+        y_t = dtype_to_ctype(dtype_z_real)
+    else:
+        y_t = out_t
+
+    # }}}
+
+    x = "x[0]" if x_is_scalar else "x[i]"
+    x = f"({x_t}) {x}"
+    y = "y[0]" if y_is_scalar else "y[i]"
+    y = f"({y_t}) {y}"
+    result = f"{x} {operator} {y}"
     return get_elwise_kernel(
         "%(tp_x)s *x, %(tp_y)s *y, %(tp_z)s *z"
         % {
@@ -496,7 +564,7 @@ def get_binary_op_kernel(dtype_x, dtype_y, dtype_z, operator):
             "tp_y": dtype_to_ctype(dtype_y),
             "tp_z": dtype_to_ctype(dtype_z),
         },
-        "z[i] = x[i] %s y[i]" % operator,
+        f"z[i] = ({out_t}) {result}",
         "multiply",
     )
 
@@ -741,14 +809,40 @@ def get_if_positive_kernel(crit_dtype, dtype):
 
 
 @context_dependent_memoize
-def get_scalar_op_kernel(dtype_x, dtype_y, operator):
+def get_where_kernel(crit_dtype, dtype):
+    return get_elwise_kernel(
+        [
+            VectorArg(crit_dtype, "crit"),
+            VectorArg(dtype, "then_"),
+            VectorArg(dtype, "else_"),
+            VectorArg(dtype, "result"),
+        ],
+        "result[i] = crit[i] != 0 ? then_[i] : else_[i]",
+        "if_positive",
+    )
+
+
+@context_dependent_memoize
+def get_scalar_op_kernel(dtype_x, dtype_a, dtype_y, operator):
     return get_elwise_kernel(
         "%(tp_x)s *x, %(tp_a)s a, %(tp_y)s *y"
         % {
             "tp_x": dtype_to_ctype(dtype_x),
             "tp_y": dtype_to_ctype(dtype_y),
-            "tp_a": dtype_to_ctype(dtype_x),
+            "tp_a": dtype_to_ctype(dtype_a),
         },
         "y[i] = x[i] %s a" % operator,
         "scalarop_kernel",
     )
+
+
+@context_dependent_memoize
+def get_logical_not_kernel(dtype_x, dtype_out):
+    return get_elwise_kernel(
+        [
+            VectorArg(dtype_x, "x"),
+            VectorArg(dtype_out, "out"),
+        ],
+        "out[i] = (x[i] == 0)",
+        "logical_not",
+    )
