@@ -25,6 +25,18 @@ def _get_common_dtype(obj1, obj2):
     return _get_common_dtype_base(obj1, obj2, has_double_support())
 
 
+def _get_broadcasted_binary_op_result(obj1, obj2,
+                                      dtype_getter=_get_common_dtype):
+
+    if obj1.shape == obj2.shape:
+        return obj1._new_like_me(dtype_getter(obj1, obj2))
+    elif obj1.shape == ():
+        return obj2._new_like_me(dtype_getter(obj1, obj2))
+    elif obj2.shape == ():
+        return obj1._new_like_me(dtype_getter(obj1, obj2))
+    else:
+        raise NotImplementedError("Broadcasting binary operator with shapes:"
+                                  f" {obj1.shape}, {obj2.shape}.")
 # {{{ vector types
 
 
@@ -141,20 +153,22 @@ def _make_binary_op(operator):
             raise RuntimeError(
                 "only contiguous arrays may " "be used as arguments to this operation"
             )
-
-        if isinstance(other, GPUArray):
-            assert self.shape == other.shape
-
+        if isinstance(other, GPUArray) and (self, GPUArray):
             if not other.flags.forc:
                 raise RuntimeError(
                     "only contiguous arrays may "
                     "be used as arguments to this operation"
                 )
 
-            result = self._new_like_me()
+            result = _get_broadcasted_binary_op_result(self, other)
             func = elementwise.get_binary_op_kernel(
-                self.dtype, other.dtype, result.dtype, operator
-            )
+                self.dtype,
+                other.dtype,
+                result.dtype,
+                operator,
+                x_is_scalar=(self.shape == ()),
+                y_is_scalar=(other.shape == ()))
+
             func.prepared_async_call(
                 self._grid,
                 self._block,
@@ -166,9 +180,12 @@ def _make_binary_op(operator):
             )
 
             return result
-        else:  # scalar operator
+        elif isinstance(self, GPUArray):  # scalar operator
+            assert np.isscalar(other)
             result = self._new_like_me()
-            func = elementwise.get_scalar_op_kernel(self.dtype, result.dtype, operator)
+            func = elementwise.get_scalar_op_kernel(self.dtype,
+                                                    np.dtype(type(other)),
+                                                    result.dtype, operator)
             func.prepared_async_call(
                 self._grid,
                 self._block,
@@ -179,6 +196,8 @@ def _make_binary_op(operator):
                 self.mem_size,
             )
             return result
+        else:
+            return AssertionError
 
     return func
 
@@ -211,7 +230,7 @@ class GPUArray:
         except TypeError:
             # handle dim-0 ndarrays:
             if isinstance(shape, np.ndarray):
-                shape = np.asscalar(shape)
+                shape = shape.item()
             assert isinstance(shape, numbers.Integral)
             s = shape
             shape = (shape,)
@@ -221,7 +240,7 @@ class GPUArray:
 
         if isinstance(s, np.integer):
             # bombs if s is a Python integer
-            s = np.asscalar(s)
+            s = s.item()
 
         if strides is None:
             if order == "F":
@@ -383,6 +402,13 @@ class GPUArray:
     def __hash__(self):
         raise TypeError("GPUArrays are not hashable.")
 
+    def __bool__(self):
+        if self.size == 1:
+            return bool(self.get())
+        else:
+            raise ValueError("The truth value of an array with "
+                    "more than one element is ambiguous. Use a.any() or a.all()")
+
     @property
     def ptr(self):
         return self.gpudata.__int__()
@@ -391,38 +417,41 @@ class GPUArray:
     def _axpbyz(self, selffac, other, otherfac, out, add_timer=None, stream=None):
         """Compute ``out = selffac * self + otherfac*other``,
         where `other` is a vector.."""
-        assert self.shape == other.shape
         if not self.flags.forc or not other.flags.forc:
             raise RuntimeError(
                 "only contiguous arrays may " "be used as arguments to this operation"
             )
-
-        func = elementwise.get_axpbyz_kernel(self.dtype, other.dtype, out.dtype)
-
+        assert ((self.shape == other.shape == out.shape)
+            or ((self.shape == ()) and other.shape == out.shape)
+            or ((other.shape == ()) and self.shape == out.shape))
+        func = elementwise.get_axpbyz_kernel(
+            self.dtype, other.dtype, out.dtype,
+            x_is_scalar=(self.shape == ()),
+            y_is_scalar=(other.shape == ()))
         if add_timer is not None:
             add_timer(
                 3 * self.size,
                 func.prepared_timed_call(
-                    self._grid,
+                    out._grid,
                     selffac,
-                    self.gpudata,
+                    out.gpudata,
                     otherfac,
                     other.gpudata,
                     out.gpudata,
-                    self.mem_size,
+                    out.mem_size,
                 ),
             )
         else:
             func.prepared_async_call(
-                self._grid,
-                self._block,
+                out._grid,
+                out._block,
                 stream,
                 selffac,
                 self.gpudata,
                 otherfac,
                 other.gpudata,
                 out.gpudata,
-                self.mem_size,
+                out.mem_size,
             )
 
         return out
@@ -454,16 +483,26 @@ class GPUArray:
             raise RuntimeError(
                 "only contiguous arrays may " "be used as arguments to this operation"
             )
+        assert ((self.shape == other.shape == out.shape)
+            or ((self.shape == ()) and other.shape == out.shape)
+            or ((other.shape == ()) and self.shape == out.shape))
+
+        func = elementwise.get_binary_op_kernel(
+            self.dtype,
+            other.dtype,
+            out.dtype,
+            "*",
+            x_is_scalar=(self.shape == ()),
+            y_is_scalar=(other.shape == ()))
 
-        func = elementwise.get_binary_op_kernel(self.dtype, other.dtype, out.dtype, "*")
         func.prepared_async_call(
-            self._grid,
-            self._block,
+            out._grid,
+            out._block,
             stream,
             self.gpudata,
             other.gpudata,
             out.gpudata,
-            self.mem_size,
+            out.mem_size,
         )
 
         return out
@@ -500,17 +539,25 @@ class GPUArray:
                 "only contiguous arrays may " "be used as arguments to this operation"
             )
 
-        assert self.shape == other.shape
+        assert ((self.shape == other.shape == out.shape)
+            or ((self.shape == ()) and other.shape == out.shape)
+            or ((other.shape == ()) and self.shape == out.shape))
 
-        func = elementwise.get_binary_op_kernel(self.dtype, other.dtype, out.dtype, "/")
+        func = elementwise.get_binary_op_kernel(
+            self.dtype,
+            other.dtype,
+            out.dtype,
+            "/",
+            x_is_scalar=(self.shape == ()),
+            y_is_scalar=(other.shape == ()))
         func.prepared_async_call(
-            self._grid,
-            self._block,
+            out._grid,
+            out._block,
             stream,
             self.gpudata,
             other.gpudata,
             out.gpudata,
-            self.mem_size,
+            out.mem_size,
         )
 
         return out
@@ -537,31 +584,35 @@ class GPUArray:
 
         if isinstance(other, GPUArray):
             # add another vector
-            result = self._new_like_me(_get_common_dtype(self, other))
+            result = _get_broadcasted_binary_op_result(self, other)
             return self._axpbyz(1, other, 1, result)
-        else:
+
+        elif np.isscalar(other):
             # add a scalar
             if other == 0:
                 return self.copy()
             else:
                 result = self._new_like_me(_get_common_dtype(self, other))
                 return self._axpbz(1, other, result)
-
+        else:
+            return NotImplemented
     __radd__ = __add__
 
     def __sub__(self, other):
         """Substract an array from an array or a scalar from an array."""
 
         if isinstance(other, GPUArray):
-            result = self._new_like_me(_get_common_dtype(self, other))
+            result = _get_broadcasted_binary_op_result(self, other)
             return self._axpbyz(1, other, -1, result)
-        else:
+        elif np.isscalar(other):
             if other == 0:
                 return self.copy()
             else:
                 # create a new array for the result
                 result = self._new_like_me(_get_common_dtype(self, other))
                 return self._axpbz(1, -other, result)
+        else:
+            return NotImplemented
 
     def __rsub__(self, other):
         """Substracts an array by a scalar or an array::
@@ -584,17 +635,22 @@ class GPUArray:
         else:
             return self._axpbz(1, -other, self)
 
+    def __pos__(self):
+        return self
+
     def __neg__(self):
         result = self._new_like_me()
         return self._axpbz(-1, 0, result)
 
     def __mul__(self, other):
         if isinstance(other, GPUArray):
-            result = self._new_like_me(_get_common_dtype(self, other))
+            result = _get_broadcasted_binary_op_result(self, other)
             return self._elwise_multiply(other, result)
-        else:
+        elif np.isscalar(other):
             result = self._new_like_me(_get_common_dtype(self, other))
             return self._axpbz(other, 0, result)
+        else:
+            return NotImplemented
 
     def __rmul__(self, scalar):
         result = self._new_like_me(_get_common_dtype(self, scalar))
@@ -612,16 +668,17 @@ class GPUArray:
         x = self / n
         """
         if isinstance(other, GPUArray):
-            result = self._new_like_me(_get_common_dtype(self, other))
+            result = _get_broadcasted_binary_op_result(self, other)
             return self._div(other, result)
-        else:
+        elif np.isscalar(other):
             if other == 1:
                 return self.copy()
             else:
                 # create a new array for the result
                 result = self._new_like_me(_get_common_dtype(self, other))
                 return self._axpbz(1 / other, 0, result)
-
+        else:
+            return NotImplemented
     __truediv__ = __div__
 
     def __rdiv__(self, other):
@@ -727,7 +784,7 @@ class GPUArray:
         if len(self.shape):
             return self.shape[0]
         else:
-            return TypeError("scalar has no len()")
+            raise TypeError("len() of unsized object")
 
     def __abs__(self):
         """Return a `GPUArray` of the absolute values of the elements
@@ -764,59 +821,41 @@ class GPUArray:
         Do the pow operator.
         with new, the user can choose between ipow or just pow
         """
-
-        if isinstance(other, GPUArray):
-            if not self.flags.forc or not other.flags.forc:
-                raise RuntimeError(
-                    "only contiguous arrays may "
-                    "be used as arguments to this operation"
-                )
-
-            assert self.shape == other.shape
-
-            if new:
-                result = self._new_like_me(_get_common_dtype(self, other))
-            else:
-                result = self
-
-            func = elementwise.get_pow_array_kernel(
-                self.dtype, other.dtype, result.dtype
-            )
-
-            func.prepared_async_call(
-                self._grid,
-                self._block,
-                None,
-                self.gpudata,
-                other.gpudata,
-                result.gpudata,
-                self.mem_size,
-            )
-
-            return result
+        common_dtype = _get_common_dtype(self, other)
+        if new:
+            result = self._new_like_me(common_dtype)
         else:
-            if not self.flags.forc:
-                raise RuntimeError(
-                    "only contiguous arrays may "
-                    "be used as arguments to this operation"
-                )
+            result = self
 
-            if new:
-                result = self._new_like_me()
-            else:
-                result = self
-            func = elementwise.get_pow_kernel(self.dtype)
-            func.prepared_async_call(
-                self._grid,
-                self._block,
-                None,
-                other,
-                self.gpudata,
-                result.gpudata,
-                self.mem_size,
-            )
+        # {{{ sanity checks
 
-            return result
+        if (not self.flags.forc) or (isinstance(other, GPUArray)
+                                     and not other.flags.forc):
+            raise RuntimeError("only contiguous arrays may "
+                               "be used as arguments to this operation")
+        assert not isinstance(other, GPUArray) or other.shape == self.shape
+
+        # }}}
+
+        func = elementwise.get_pow_array_kernel(
+            self.dtype,
+            common_dtype if np.isscalar(other) else other.dtype,
+            result.dtype,
+            not np.isscalar(self),
+            not np.isscalar(other)
+        )
+
+        func.prepared_async_call(
+            self._grid,
+            self._block,
+            None,
+            self.gpudata,
+            other if np.isscalar(other) else other.gpudata,
+            result.gpudata,
+            self.mem_size,
+        )
+
+        return result
 
     def __pow__(self, other):
         """pow function::
@@ -839,6 +878,27 @@ class GPUArray:
         """
         return self._pow(other, new=False)
 
+    def __rpow__(self, other):
+        common_dtype = _get_common_dtype(self, other)
+        result = self._new_like_me(common_dtype)
+
+        if not np.isscalar(other):
+            # Base is a gpuarray => do not cast.
+            base = other
+        else:
+            base = common_dtype.type(other)
+
+        func = elementwise.get_pow_array_kernel(
+            base.dtype, self.dtype, result.dtype,
+            is_base_array=not np.isscalar(other), is_exp_array=not np.isscalar(self))
+        # Evaluates z = x ** y
+        func.prepared_async_call(self._grid, self._block, None,
+                                 base if np.isscalar(base) else base.gpudata,  # x
+                                 self.gpudata,  # y
+                                 result.gpudata,  # z
+                                 self.mem_size)
+        return result
+
     def reverse(self, stream=None):
         """Return this array in reversed order. The array is treated
         as one-dimensional.
@@ -876,6 +936,12 @@ class GPUArray:
 
         return result
 
+    def any(self, stream=None, allocator=None):
+        return any(self, stream=stream, allocator=allocator)
+
+    def all(self, stream=None, allocator=None):
+        return all(self, stream=stream, allocator=allocator)
+
     def reshape(self, *shape, **kwargs):
         """Gives a new shape to an array without changing its data."""
 
@@ -1000,7 +1066,7 @@ class GPUArray:
             shape=tuple(new_shape),
             dtype=self.dtype,
             allocator=self.allocator,
-            base=self.base or self,
+            base=self if self.base is None else self.base,
             gpudata=self.gpudata,
             strides=tuple(new_strides),
         )
@@ -1100,7 +1166,10 @@ class GPUArray:
         )
 
     def __setitem__(self, index, value):
-        _memcpy_discontig(self[index], value)
+        if np.isscalar(value):
+            self[index].fill(value)
+        else:
+            _memcpy_discontig(self[index], value)
 
     # }}}
 
@@ -1237,7 +1306,7 @@ def to_gpu_async(ary, allocator=drv.mem_alloc, stream=None):
 empty = GPUArray
 
 
-def zeros(shape, dtype, allocator=drv.mem_alloc, order="C"):
+def zeros(shape, dtype=np.float64, allocator=drv.mem_alloc, order="C"):
     """Returns an array of the given shape and dtype filled with 0's."""
     result = GPUArray(shape, dtype, allocator, order=order)
     zero = np.zeros((), dtype)
@@ -1245,6 +1314,14 @@ def zeros(shape, dtype, allocator=drv.mem_alloc, order="C"):
     return result
 
 
+def ones(shape, dtype=np.float64, allocator=drv.mem_alloc, order="C"):
+    """Returns an array of the given shape and dtype filled with 1's."""
+    result = GPUArray(shape, dtype, allocator, order=order)
+    one = np.ones((), dtype)
+    result.fill(one)
+    return result
+
+
 def _array_like_helper(other_ary, dtype, order):
     """Set order, strides, dtype as in numpy's zero_like. """
     strides = None
@@ -1767,6 +1844,96 @@ def multi_put(arrays, dest_indices, dest_shape=None, out=None, stream=None):
 
 # {{{ shape manipulation
 
+def concatenate(arrays, axis=0, allocator=None):
+    """
+    Join a sequence of arrays along an existing axis.
+    :arg arrays: A sequnce of :class:`GPUArray`.
+    :arg axis: Index of the dimension of the new axis in the result array.
+        Can be -1, for the new axis to be last dimension.
+    :returns: :class:`GPUArray`
+    """
+    # implementation is borrowed from pyopencl.array.concatenate()
+    # {{{ find properties of result array
+
+    shape = None
+
+    def shape_except_axis(ary: GPUArray):
+        return ary.shape[:axis] + ary.shape[axis+1:]
+
+    for i_ary, ary in enumerate(arrays):
+        allocator = allocator or ary.allocator
+
+        if shape is None:
+            # first array
+            shape = list(ary.shape)
+
+        else:
+            if len(ary.shape) != len(shape):
+                raise ValueError("%d'th array has different number of axes "
+                        "(should have %d, has %d)"
+                        % (i_ary, len(ary.shape), len(shape)))
+
+            if (ary.ndim != arrays[0].ndim
+                    or shape_except_axis(ary) != shape_except_axis(arrays[0])):
+                raise ValueError("%d'th array has residual not matching "
+                        "other arrays" % i_ary)
+
+            shape[axis] += ary.shape[axis]
+
+    # }}}
+
+    shape = tuple(shape)
+    dtype = np.find_common_type([ary.dtype for ary in arrays], [])
+    result = empty(shape, dtype, allocator=allocator)
+
+    full_slice = (slice(None),) * len(shape)
+
+    base_idx = 0
+    for ary in arrays:
+        my_len = ary.shape[axis]
+        result[full_slice[:axis] + (slice(base_idx, base_idx+my_len),) + full_slice[axis+1:]] = ary
+        base_idx += my_len
+
+    return result
+
+
+def stack(arrays, axis=0, allocator=None):
+    """
+    Join a sequence of arrays along a new axis.
+    :arg arrays: A sequnce of :class:`GPUArray`.
+    :arg axis: Index of the dimension of the new axis in the result array.
+        Can be -1, for the new axis to be last dimension.
+    :returns: :class:`GPUArray`
+    """
+    # implementation is borrowed from pyopencl.array.stack()
+    allocator = allocator or arrays[0].allocator
+
+    if not arrays:
+        raise ValueError("need at least one array to stack")
+
+    input_shape = arrays[0].shape
+    input_ndim = arrays[0].ndim
+    axis = input_ndim if axis == -1 else axis
+
+    import builtins
+    if not builtins.all(ary.shape == input_shape for ary in arrays[1:]):
+        raise ValueError("arrays must have the same shape")
+
+    if not (0 <= axis <= input_ndim):
+        raise ValueError("invalid axis")
+
+    result_shape = input_shape[:axis] + (len(arrays),) + input_shape[axis:]
+    result = empty(shape=result_shape,
+            dtype=np.result_type(*(ary.dtype for ary in arrays)),
+            allocator=allocator, order="C" if axis == 0 else "F")
+
+    for i, ary in enumerate(arrays):
+
+        idx = (slice(None),)*axis + (i,) + (slice(None),)*(input_ndim-axis)
+        result[idx] = ary
+
+    return result
+
 
 def transpose(a, axes=None):
     """Permute the dimensions of an array.
@@ -1824,6 +1991,32 @@ def if_positive(criterion, then_, else_, out=None, stream=None):
     return out
 
 
+def where(criterion, then_, else_, out=None, stream=None):
+    if (criterion.shape != then_.shape != else_.shape):
+        raise NotImplementedError("shape broadcast not implemented")
+
+    if (then_.dtype != else_.dtype):
+        raise NotImplementedError("dtype broadcast not implemented")
+
+    func = elementwise.get_where_kernel(criterion.dtype, then_.dtype)
+
+    if out is None:
+        out = empty_like(then_)
+
+    func.prepared_async_call(
+        criterion._grid,
+        criterion._block,
+        stream,
+        criterion.gpudata,
+        then_.gpudata,
+        else_.gpudata,
+        out.gpudata,
+        criterion.size,
+    )
+
+    return out
+
+
 def _make_binary_minmax_func(which):
     def f(a, b, out=None, stream=None):
         if isinstance(a, GPUArray) and isinstance(b, GPUArray):
@@ -1876,6 +2069,20 @@ def sum(a, dtype=None, stream=None, allocator=None):
     return krnl(a, stream=stream, allocator=allocator)
 
 
+def any(a, stream=None, allocator=None):
+    from pycuda.reduction import get_any_kernel
+
+    krnl = get_any_kernel(np.dtype(bool), a.dtype)
+    return krnl(a, stream=stream, allocator=allocator)
+
+
+def all(a, stream=None, allocator=None):
+    from pycuda.reduction import get_all_kernel
+
+    krnl = get_all_kernel(np.dtype(bool), a.dtype)
+    return krnl(a, stream=stream, allocator=allocator)
+
+
 def subset_sum(subset, a, dtype=None, stream=None, allocator=None):
     from pycuda.reduction import get_subset_sum_kernel
 
@@ -1930,4 +2137,95 @@ subset_max = _make_subset_minmax_kernel("max")
 
 # }}}
 
+
+# {{{ logical ops
+
+def _logical_op(x1, x2, out, allocator, operator):
+    assert operator in ["&&", "||"]
+    allocator = (
+        allocator
+        or getattr(x1, "allocator", None)
+        or getattr(x2, "allocator", None)
+        or drv.mem_alloc)
+
+    if np.isscalar(x1) and np.isscalar(x2):
+        if out is None:
+            out = empty(shape=(), dtype=np.bool_, allocator=allocator)
+
+        if operator == "&&":
+            out[:] = np.logical_and(x1, x2)
+        else:
+            out[:] = np.logical_or(x1, x2)
+    elif np.isscalar(x1) or np.isscalar(x2):
+        scalar_arg, = [x for x in (x1, x2) if np.isscalar(x)]
+        ary_arg, = [x for x in (x1, x2) if not np.isscalar(x)]
+        if not isinstance(ary_arg, GPUArray):
+            raise ValueError("logical_and can take either scalar or GPUArrays"
+                             " as inputs")
+
+        out = out or ary_arg._new_like_me(dtype=np.bool_)
+
+        assert out.shape == ary_arg.shape and out.dtype == np.bool_
+
+        func = elementwise.get_scalar_op_kernel(ary_arg.dtype,
+                                                np.dtype(type(scalar_arg)),
+                                                out.dtype,
+                                                operator)
+
+        func.prepared_async_call(out._grid, out._block,
+                                 None,
+                                 ary_arg.gpudata,
+                                 scalar_arg,
+                                 out.gpudata,
+                                 out.mem_size)
+    else:
+        if not (isinstance(x1, GPUArray) and isinstance(x2, GPUArray)):
+            raise ValueError("logical_and can take either scalar or GPUArrays"
+                             " as inputs")
+        if x1.shape != x2.shape:
+            raise NotImplementedError("Broadcasting not supported")
+
+        if out is None:
+            out = x1._new_like_me(dtype=np.bool_)
+
+        assert out.shape == x1.shape and out.dtype == np.bool_
+
+        func = elementwise.get_binary_op_kernel(
+            x1.dtype, x2.dtype, out.dtype, operator
+        )
+        func.prepared_async_call(out._grid, out._block,
+                                 None,
+                                 x1.gpudata,
+                                 x2.gpudata,
+                                 out.gpudata,
+                                 out.mem_size)
+
+    return out
+
+
+def logical_and(x1, x2, /, out=None, *, allocator=None):
+    return _logical_op(x1, x2, out, allocator, "&&")
+
+
+def logical_or(x1, x2, /, out=None, *, allocator=None):
+    return _logical_op(x1, x2, out, allocator, "||")
+
+
+def logical_not(x, /, out=None, *, allocator=drv.mem_alloc):
+    if np.isscalar(x):
+        out = out or empty(shape=(), dtype=np.bool_, allocator=allocator)
+        out[:] = np.logical_not(x)
+    else:
+        out = out or empty(shape=x.shape, dtype=np.bool_, allocator=allocator)
+        func = elementwise.get_logical_not_kernel(x.dtype, out.dtype)
+        func.prepared_async_call(out._grid, out._block,
+                                 None,
+                                 x.gpudata,
+                                 out.gpudata,
+                                 out.mem_size)
+
+    return out
+
+# }}}
+
 # vim: foldmethod=marker
