@@ -1030,11 +1030,38 @@ BOOST_PYTHON_MODULE(_driver)
     .value("COMPUTE_12", CU_TARGET_COMPUTE_12)
     .value("COMPUTE_13", CU_TARGET_COMPUTE_13)
 #endif
-#if CUDAPP_CUDA_VERSION >= 3000
+#if (CUDAPP_CUDA_VERSION >= 3000) && CUDAPP_CUDA_VERSION < 12000
     .value("COMPUTE_20", CU_TARGET_COMPUTE_20)
 #endif
-#if CUDAPP_CUDA_VERSION >= 3020
+#if (CUDAPP_CUDA_VERSION >= 3020) && CUDAPP_CUDA_VERSION < 12000
     .value("COMPUTE_21", CU_TARGET_COMPUTE_21)
+#endif
+#if (CUDAPP_CUDA_VERSION >= 11000)
+    .value("COMPUTE_30", CU_TARGET_COMPUTE_30)
+    .value("COMPUTE_32", CU_TARGET_COMPUTE_32)
+    .value("COMPUTE_35", CU_TARGET_COMPUTE_35)
+    .value("COMPUTE_37", CU_TARGET_COMPUTE_37)
+
+    .value("COMPUTE_50", CU_TARGET_COMPUTE_50)
+    .value("COMPUTE_52", CU_TARGET_COMPUTE_52)
+    .value("COMPUTE_53", CU_TARGET_COMPUTE_53)
+
+    .value("COMPUTE_60", CU_TARGET_COMPUTE_60)
+    .value("COMPUTE_61", CU_TARGET_COMPUTE_61)
+    .value("COMPUTE_62", CU_TARGET_COMPUTE_62)
+
+    .value("COMPUTE_70", CU_TARGET_COMPUTE_70)
+    .value("COMPUTE_72", CU_TARGET_COMPUTE_72)
+    .value("COMPUTE_75", CU_TARGET_COMPUTE_75)
+
+    .value("COMPUTE_80", CU_TARGET_COMPUTE_80)
+#endif
+#if (CUDAPP_CUDA_VERSION >= 11060) /* approximate, precise version unknown */
+    .value("COMPUTE_86", CU_TARGET_COMPUTE_86)
+#endif
+#if (CUDAPP_CUDA_VERSION >= 12000)
+    .value("COMPUTE_90", CU_TARGET_COMPUTE_90)
+    .value("COMPUTE_90A", CU_TARGET_COMPUTE_90A)
 #endif
     ;
 
