@@ -3,17 +3,47 @@
 import numpy as np
 import numpy.linalg as la
 import sys
-from pycuda.tools import mark_cuda_test
+from pycuda.tools import init_cuda_context_fixture
 from pycuda.characterize import has_double_support
 
 
 import pycuda.gpuarray as gpuarray
 import pycuda.driver as drv
 from pycuda.compiler import SourceModule
+import pytest
+import operator
 
 
+@pytest.fixture(autouse=True)
+def init_cuda_context():
+    yield from init_cuda_context_fixture()
+
+
+def get_random_array(rng, shape, dtype):
+    dtype = np.dtype(dtype)
+
+    if dtype.kind == "f":
+        return rng.random(shape, dtype)
+    elif dtype.kind in "il":
+        return rng.integers(-42, 42, shape, dtype)
+    elif dtype.kind in "u":
+        return rng.integers(0, 42, shape, dtype)
+    elif dtype.kind == "c":
+        real_dtype = np.empty(0, dtype).real.dtype
+        return (dtype.type(1j) * get_random_array(rng, shape, real_dtype)
+                + get_random_array(rng, shape, real_dtype))
+    else:
+        raise NotImplementedError(f"dtype = {dtype}")
+
+
+def skip_if_not_enough_gpu_memory(required_mem_gigabytes):
+    device_mem_GB = drv.Context.get_device().total_memory() / 1e9
+    if device_mem_GB < required_mem_gigabytes:
+        pytest.skip("Need at least %.1f GB memory" % required_mem_gigabytes)
+
+
+@pytest.mark.cuda
 class TestGPUArray:
-    @mark_cuda_test
     def test_pow_array(self):
         a = np.array([1, 2, 3, 4, 5]).astype(np.float32)
         a_gpu = gpuarray.to_gpu(a)
@@ -30,9 +60,9 @@ class TestGPUArray:
         a_gpu = a_gpu.get()
         np.testing.assert_allclose(pow(a, b), a_gpu, rtol=1e-6)
 
-    @mark_cuda_test
-    def test_pow_number(self):
-        a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(np.float32)
+    @pytest.mark.parametrize("dtype", [np.float32, np.float64])
+    def test_pow_number(self, dtype):
+        a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(dtype)
         a_gpu = gpuarray.to_gpu(a)
 
         result = pow(a_gpu, 2).get()
@@ -42,7 +72,6 @@ class TestGPUArray:
         a_gpu = a_gpu.get()
         np.testing.assert_allclose(a ** 2, a_gpu, rtol=1e-6)
 
-    @mark_cuda_test
     def test_rpow_array(self):
         scalar = np.random.rand()
         a = abs(np.random.rand(10))
@@ -57,18 +86,15 @@ class TestGPUArray:
         result = (a_gpu ** scalar).get()
         np.testing.assert_allclose(a ** scalar, result)
 
-    @mark_cuda_test
     def test_numpy_integer_shape(self):
         gpuarray.empty(np.int32(17), np.float32)
         gpuarray.empty((np.int32(17), np.int32(17)), np.float32)
 
-    @mark_cuda_test
     def test_ndarray_shape(self):
         gpuarray.empty(np.array(3), np.float32)
         gpuarray.empty(np.array([3]), np.float32)
         gpuarray.empty(np.array([2, 3]), np.float32)
 
-    @mark_cuda_test
     def test_abs(self):
         a = -gpuarray.arange(111, dtype=np.float32)
         res = a.get()
@@ -84,13 +110,11 @@ class TestGPUArray:
             assert abs(res[i]) >= 0
             assert res[i] == i
 
-    @mark_cuda_test
     def test_len(self):
         a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(np.float32)
         a_cpu = gpuarray.to_gpu(a)
         assert len(a_cpu) == 10
 
-    @mark_cuda_test
     def test_multiply(self):
         """Test the muliplication of an array with a scalar. """
 
@@ -103,7 +127,6 @@ class TestGPUArray:
 
                     assert (a * scalar == a_doubled).all()
 
-    @mark_cuda_test
     def test_rmul_yields_right_type(self):
         a = np.array([1, 2, 3, 4, 5]).astype(np.float32)
         a_gpu = gpuarray.to_gpu(a)
@@ -114,30 +137,56 @@ class TestGPUArray:
         two_a = np.float32(2) * a_gpu
         assert isinstance(two_a, gpuarray.GPUArray)
 
-    @mark_cuda_test
     def test_multiply_array(self):
         """Test the multiplication of two arrays."""
 
         a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(np.float32)
+        b = np.array([10, 20, 30, 40, 50, 60, 70, 80, 90, 100]).astype(np.float32)
+        c = np.array(2)
 
         a_gpu = gpuarray.to_gpu(a)
-        b_gpu = gpuarray.to_gpu(a)
+        b_gpu = gpuarray.to_gpu(b)
+        c_gpu = gpuarray.to_gpu(c)
 
-        a_squared = (b_gpu * a_gpu).get()
+        a_mul_b = (a_gpu * b_gpu).get()
+        assert (a * b == a_mul_b).all()
 
-        assert (a * a == a_squared).all()
+        b_mul_a = (b_gpu * a_gpu).get()
+        assert (b * a == b_mul_a).all()
+
+        a_mul_c = (a_gpu * c_gpu).get()
+        assert (a * c == a_mul_c).all()
+
+        b_mul_c = (b_gpu * c_gpu).get()
+        assert (b * c == b_mul_c).all()
+
+    def test_unit_multiply_array(self):
+
+        a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(np.float32)
+
+        a_gpu = gpuarray.to_gpu(a)
+        np.testing.assert_allclose(+a_gpu.get(), +a, rtol=1e-6)
+        np.testing.assert_allclose(-a_gpu.get(), -a, rtol=1e-6)
 
-    @mark_cuda_test
     def test_addition_array(self):
         """Test the addition of two arrays."""
 
         a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(np.float32)
         a_gpu = gpuarray.to_gpu(a)
+        b = np.array(1).astype(np.float32)
+        b_gpu = gpuarray.to_gpu(b)
         a_added = (a_gpu + a_gpu).get()
+        a_added_scalar = (a_gpu + 1).get()
+        scalar_added_a = (1 + a_gpu).get()
+        a_gpu_pl_b_gpu = (a_gpu + b_gpu).get()
+        b_gpu_pl_a_gpu = (b_gpu + a_gpu).get()
 
         assert (a + a == a_added).all()
+        assert (a + 1 == a_added_scalar).all()
+        assert (1 + a == scalar_added_a).all()
+        assert (a + b == a_gpu_pl_b_gpu).all()
+        assert (b + a == b_gpu_pl_a_gpu).all()
 
-    @mark_cuda_test
     def test_iaddition_array(self):
         """Test the inplace addition of two arrays."""
 
@@ -148,7 +197,6 @@ class TestGPUArray:
 
         assert (a + a == a_added).all()
 
-    @mark_cuda_test
     def test_addition_scalar(self):
         """Test the addition of an array and a scalar."""
 
@@ -158,7 +206,6 @@ class TestGPUArray:
 
         assert (7 + a == a_added).all()
 
-    @mark_cuda_test
     def test_iaddition_scalar(self):
         """Test the inplace addition of an array and a scalar."""
 
@@ -169,15 +216,16 @@ class TestGPUArray:
 
         assert (7 + a == a_added).all()
 
-    @mark_cuda_test
     def test_substract_array(self):
         """Test the subtraction of two arrays."""
         # test data
         a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(np.float32)
         b = np.array([10, 20, 30, 40, 50, 60, 70, 80, 90, 100]).astype(np.float32)
+        c = np.array(1).astype(np.float32)
 
         a_gpu = gpuarray.to_gpu(a)
         b_gpu = gpuarray.to_gpu(b)
+        c_gpu = gpuarray.to_gpu(c)
 
         result = (a_gpu - b_gpu).get()
         assert (a - b == result).all()
@@ -185,7 +233,12 @@ class TestGPUArray:
         result = (b_gpu - a_gpu).get()
         assert (b - a == result).all()
 
-    @mark_cuda_test
+        result = (a_gpu - c_gpu).get()
+        assert (a - c == result).all()
+
+        result = (c_gpu - a_gpu).get()
+        assert (c - a == result).all()
+
     def test_substract_scalar(self):
         """Test the subtraction of an array and a scalar."""
 
@@ -201,7 +254,6 @@ class TestGPUArray:
         result = (7 - a_gpu).get()
         assert (7 - a == result).all()
 
-    @mark_cuda_test
     def test_divide_scalar(self):
         """Test the division of an array and a scalar."""
 
@@ -214,16 +266,17 @@ class TestGPUArray:
         result = (2 / a_gpu).get()
         assert (2 / a == result).all()
 
-    @mark_cuda_test
     def test_divide_array(self):
         """Test the division of an array and a scalar. """
 
         # test data
         a = np.array([10, 20, 30, 40, 50, 60, 70, 80, 90, 100]).astype(np.float32)
         b = np.array([10, 10, 10, 10, 10, 10, 10, 10, 10, 10]).astype(np.float32)
+        c = np.array(2)
 
         a_gpu = gpuarray.to_gpu(a)
         b_gpu = gpuarray.to_gpu(b)
+        c_gpu = gpuarray.to_gpu(c)
 
         a_divide = (a_gpu / b_gpu).get()
         assert (np.abs(a / b - a_divide) < 1e-3).all()
@@ -231,7 +284,12 @@ class TestGPUArray:
         a_divide = (b_gpu / a_gpu).get()
         assert (np.abs(b / a - a_divide) < 1e-3).all()
 
-    @mark_cuda_test
+        a_divide = (a_gpu / c_gpu).get()
+        assert (np.abs(a / c - a_divide) < 1e-3).all()
+
+        a_divide = (c_gpu / a_gpu).get()
+        assert (np.abs(c / a - a_divide) < 1e-3).all()
+
     def test_random(self):
         from pycuda.curandom import rand as curand
 
@@ -246,7 +304,6 @@ class TestGPUArray:
             assert (0 <= a).all()
             assert (a < 1).all()
 
-    @mark_cuda_test
     def test_curand_wrappers(self):
         from pycuda.curandom import get_curand_version
 
@@ -321,7 +378,6 @@ class TestGPUArray:
                     # # Compare with scipy.stats.poisson.pmf(v - 1, v)
                     # assert np.isclose(0.12511, tmp, atol=0.002)
 
-    @mark_cuda_test
     def test_array_gt(self):
         """Test whether array contents are > the other array's
         contents"""
@@ -334,7 +390,6 @@ class TestGPUArray:
         assert result[0]
         assert not result[1]
 
-    @mark_cuda_test
     def test_array_lt(self):
         """Test whether array contents are < the other array's
         contents"""
@@ -347,7 +402,6 @@ class TestGPUArray:
         assert result[0]
         assert not result[1]
 
-    @mark_cuda_test
     def test_array_le(self):
         """Test whether array contents are <= the other array's
         contents"""
@@ -361,7 +415,6 @@ class TestGPUArray:
         assert result[1]
         assert not result[2]
 
-    @mark_cuda_test
     def test_array_ge(self):
         """Test whether array contents are >= the other array's
         contents"""
@@ -375,7 +428,6 @@ class TestGPUArray:
         assert result[1]
         assert not result[2]
 
-    @mark_cuda_test
     def test_array_eq(self):
         """Test whether array contents are == the other array's
         contents"""
@@ -388,7 +440,6 @@ class TestGPUArray:
         assert not result[0]
         assert result[1]
 
-    @mark_cuda_test
     def test_array_ne(self):
         """Test whether array contents are != the other array's
         contents"""
@@ -401,7 +452,6 @@ class TestGPUArray:
         assert result[0]
         assert not result[1]
 
-    @mark_cuda_test
     def test_nan_arithmetic(self):
         def make_nan_contaminated_vector(size):
             shape = (size,)
@@ -426,7 +476,6 @@ class TestGPUArray:
 
         assert (np.isnan(ab) == np.isnan(ab_gpu)).all()
 
-    @mark_cuda_test
     def test_elwise_kernel(self):
         from pycuda.curandom import rand as curand
 
@@ -446,7 +495,6 @@ class TestGPUArray:
 
         assert la.norm((c_gpu - (5 * a_gpu + 6 * b_gpu)).get()) < 1e-5
 
-    @mark_cuda_test
     def test_ranged_elwise_kernel(self):
         from pycuda.elementwise import ElementwiseKernel
 
@@ -470,7 +518,6 @@ class TestGPUArray:
 
             assert la.norm(a_cpu - a_gpu.get()) == 0, i
 
-    @mark_cuda_test
     def test_take(self):
         idx = gpuarray.arange(0, 10000, 2, dtype=np.uint32)
         for dtype in [np.float32, np.complex64]:
@@ -480,40 +527,42 @@ class TestGPUArray:
 
             assert (a_host[idx.get()] == result.get()).all()
 
-    @mark_cuda_test
     def test_arange(self):
         a = gpuarray.arange(12, dtype=np.float32)
         assert (np.arange(12, dtype=np.float32) == a.get()).all()
 
-    @mark_cuda_test
-    def test_stack(self):
+    def test_ones(self):
 
-        orders = ["F", "C"]
-        input_dims_lst = [0, 1, 2]
+        ones = np.ones(10)
+        ones_gpu = gpuarray.ones(10)
 
-        for order in orders:
-            for input_dims in input_dims_lst:
-                shape = (2, 2, 2)[:input_dims]
-                axis = -1 if order == "F" else 0
+        np.testing.assert_allclose(ones, ones_gpu.get(), rtol=1e-6)
+        assert ones.dtype == ones_gpu.dtype
 
-                from numpy.random import default_rng
-                rng = default_rng()
-                x_in = rng.random(size=shape)
-                y_in = rng.random(size=shape)
-                x_in = x_in if order == "C" else np.asfortranarray(x_in)
-                y_in = y_in if order == "C" else np.asfortranarray(y_in)
+    @pytest.mark.parametrize("order", ["F", "C"])
+    @pytest.mark.parametrize("input_dims", [0, 1, 2])
+    def test_stack(self, order, input_dims):
 
-                x_gpu = gpuarray.to_gpu(x_in)
-                y_gpu = gpuarray.to_gpu(y_in)
+        shape = (2, 2, 2)[:input_dims]
+        axis = -1 if order == "F" else 0
 
-                numpy_stack = np.stack((x_in, y_in), axis=axis)
-                gpuarray_stack = gpuarray.stack((x_gpu, y_gpu), axis=axis)
+        from numpy.random import default_rng
+        rng = default_rng()
+        x_in = rng.random(size=shape)
+        y_in = rng.random(size=shape)
+        x_in = x_in if order == "C" else np.asfortranarray(x_in)
+        y_in = y_in if order == "C" else np.asfortranarray(y_in)
 
-                np.testing.assert_allclose(gpuarray_stack.get(), numpy_stack)
+        x_gpu = gpuarray.to_gpu(x_in)
+        y_gpu = gpuarray.to_gpu(y_in)
 
-                assert gpuarray_stack.shape == numpy_stack.shape
+        numpy_stack = np.stack((x_in, y_in), axis=axis)
+        gpuarray_stack = gpuarray.stack((x_gpu, y_gpu), axis=axis)
+
+        np.testing.assert_allclose(gpuarray_stack.get(), numpy_stack)
+
+        assert gpuarray_stack.shape == numpy_stack.shape
 
-    @mark_cuda_test
     def test_concatenate(self):
 
         from pycuda.curandom import rand as curand
@@ -532,7 +581,6 @@ class TestGPUArray:
 
         assert cat.shape == cat_dev.shape
 
-    @mark_cuda_test
     def test_reverse(self):
         a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(np.float32)
         a_cpu = gpuarray.to_gpu(a)
@@ -544,7 +592,6 @@ class TestGPUArray:
         for i in range(0, 10):
             assert a[len(a) - 1 - i] == b[i]
 
-    @mark_cuda_test
     def test_sum(self):
         from pycuda.curandom import rand as curand
 
@@ -557,7 +604,56 @@ class TestGPUArray:
 
         assert abs(sum_a_gpu - sum_a) / abs(sum_a) < 1e-4
 
-    @mark_cuda_test
+    @pytest.mark.parametrize("dtype", [np.int32, np.bool_, np.float32, np.float64])
+    def test_any(self, dtype):
+
+        ary_list = [np.ones(10, dtype),
+                    np.zeros(1, dtype),
+                    np.ones(1, dtype),
+                    np.empty(10, dtype)]
+
+        for ary in ary_list:
+            ary_gpu = gpuarray.to_gpu(ary)
+            any_ary = np.any(ary)
+            any_ary_gpu = ary_gpu.any().get()
+            np.testing.assert_array_equal(any_ary_gpu, any_ary)
+            assert any_ary_gpu.dtype == any_ary.dtype
+
+        import itertools
+        for _array in list(itertools.product([0, 1], [0, 1], [0, 1])):
+            array = np.array(_array, dtype)
+            array_gpu = gpuarray.to_gpu(array)
+            any_array = np.any(array)
+            any_array_gpu = array_gpu.any().get()
+
+            np.testing.assert_array_equal(any_array_gpu, any_array)
+            assert any_array_gpu.dtype == any_array.dtype
+
+    @pytest.mark.parametrize("dtype", [np.int32, np.bool_, np.float32, np.float64])
+    def test_all(self, dtype):
+
+        ary_list = [np.ones(10, dtype),
+                    np.zeros(1, dtype),
+                    np.ones(1, dtype),
+                    np.empty(10, dtype)]
+
+        for ary in ary_list:
+            ary_gpu = gpuarray.to_gpu(ary)
+            all_ary = np.all(ary)
+            all_ary_gpu = ary_gpu.all().get()
+            np.testing.assert_array_equal(all_ary_gpu, all_ary)
+            assert all_ary_gpu.dtype == all_ary.dtype
+
+        import itertools
+        for _array in list(itertools.product([0, 1], [0, 1], [0, 1])):
+            array = np.array(_array, dtype)
+            array_gpu = gpuarray.to_gpu(array)
+            all_array = np.all(array)
+            all_array_gpu = array_gpu.all().get()
+
+            np.testing.assert_array_equal(all_array_gpu, all_array)
+            assert all_array_gpu.dtype == all_array.dtype
+
     def test_minmax(self):
         from pycuda.curandom import rand as curand
 
@@ -576,7 +672,6 @@ class TestGPUArray:
 
                 assert op_a_gpu == op_a, (op_a_gpu, op_a, dtype, what)
 
-    @mark_cuda_test
     def test_subset_minmax(self):
         from pycuda.curandom import rand as curand
 
@@ -610,41 +705,38 @@ class TestGPUArray:
 
             assert min_a_gpu == min_a
 
-    @mark_cuda_test
-    def test_dot(self):
+    @pytest.mark.parametrize("sz", [2,
+                                    3,
+                                    4,
+                                    5,
+                                    6,
+                                    7,
+                                    31,
+                                    32,
+                                    33,
+                                    127,
+                                    128,
+                                    129,
+                                    255,
+                                    256,
+                                    257,
+                                    16384 - 993,
+                                    20000,
+                                    ])
+    def test_dot(self, sz):
         from pycuda.curandom import rand as curand
 
-        for sz in [
-            2,
-            3,
-            4,
-            5,
-            6,
-            7,
-            31,
-            32,
-            33,
-            127,
-            128,
-            129,
-            255,
-            256,
-            257,
-            16384 - 993,
-            20000,
-        ]:
-            a_gpu = curand((sz,))
-            a = a_gpu.get()
-            b_gpu = curand((sz,))
-            b = b_gpu.get()
+        a_gpu = curand((sz,))
+        a = a_gpu.get()
+        b_gpu = curand((sz,))
+        b = b_gpu.get()
 
-            dot_ab = np.dot(a, b)
+        dot_ab = np.dot(a, b)
 
-            dot_ab_gpu = gpuarray.dot(a_gpu, b_gpu).get()
+        dot_ab_gpu = gpuarray.dot(a_gpu, b_gpu).get()
 
-            assert abs(dot_ab_gpu - dot_ab) / abs(dot_ab) < 1e-4
+        assert abs(dot_ab_gpu - dot_ab) / abs(dot_ab) < 1e-4
 
-    @mark_cuda_test
     def test_slice(self):
         from pycuda.curandom import rand as curand
 
@@ -663,7 +755,6 @@ class TestGPUArray:
 
             assert la.norm(a_gpu_slice.get() - a_slice) == 0
 
-    @mark_cuda_test
     def test_2d_slice_c(self):
         from pycuda.curandom import rand as curand
 
@@ -683,7 +774,6 @@ class TestGPUArray:
 
             assert la.norm(a_gpu_slice.get() - a_slice) == 0
 
-    @mark_cuda_test
     def test_2d_slice_f(self):
         from pycuda.curandom import rand as curand
         import pycuda.gpuarray as gpuarray
@@ -707,7 +797,22 @@ class TestGPUArray:
 
             assert la.norm(a_gpu_slice.get() - a_slice) == 0
 
-    @mark_cuda_test
+    def test_where(self):
+        a = np.array([1, 0, -1])
+        b = np.array([2, 2, 2])
+        c = np.array([3, 3, 3])
+
+        import pycuda.gpuarray as gpuarray
+
+        a_gpu = gpuarray.to_gpu(a)
+        b_gpu = gpuarray.to_gpu(b)
+        c_gpu = gpuarray.to_gpu(c)
+
+        result = gpuarray.where(a_gpu, b_gpu, c_gpu).get()
+        result_ref = np.where(a, b, c)
+
+        np.testing.assert_allclose(result_ref, result, rtol=1e-5)
+
     def test_if_positive(self):
         from pycuda.curandom import rand as curand
 
@@ -728,7 +833,6 @@ class TestGPUArray:
         assert la.norm(max_a_b_gpu.get() - np.maximum(a, b)) == 0
         assert la.norm(min_a_b_gpu.get() - np.minimum(a, b)) == 0
 
-    @mark_cuda_test
     def test_take_put(self):
         for n in [5, 17, 333]:
             one_field_size = 8
@@ -750,7 +854,6 @@ class TestGPUArray:
 
             drv.Context.synchronize()
 
-    @mark_cuda_test
     def test_astype(self):
         from pycuda.curandom import rand as curand
 
@@ -773,7 +876,6 @@ class TestGPUArray:
         assert a2.dtype == np.float32
         assert la.norm(a - a2) / la.norm(a) < 1e-7
 
-    @mark_cuda_test
     def test_complex_bits(self):
         from pycuda.curandom import rand as curand
 
@@ -799,7 +901,7 @@ class TestGPUArray:
             # verify conj with out parameter
             z_out = z.astype(np.complex64)
             assert z_out is z.conj(out=z_out)
-            assert la.norm(z.get().conj() - z_out.get()) < 1e-7
+            assert la.norm(z.get().conj() - z_out.get()) < 5e-6
 
             # verify contiguity is preserved
             for order in ["C", "F"]:
@@ -818,7 +920,6 @@ class TestGPUArray:
                         assert zdata.imag.flags.f_contiguous
                         assert zdata.conj().flags.f_contiguous
 
-    @mark_cuda_test
     def test_pass_slice_to_kernel(self):
         mod = SourceModule(
             """
@@ -839,9 +940,9 @@ class TestGPUArray:
 
         a = a_gpu.get()
         assert (a[255:257] == np.array([1, 2], np.float32)).all()
-        assert (a[255 * 256 - 1: 255 * 256 + 1] == np.array([2, 1], np.float32)).all()
+        np.testing.assert_array_equal(a[255 * 256 - 1: 255 * 256 + 1],
+                                      np.array([2, 1], np.float32))
 
-    @mark_cuda_test
     def test_scan(self):
         from pycuda.scan import ExclusiveScanKernel, InclusiveScanKernel
 
@@ -870,7 +971,6 @@ class TestGPUArray:
 
                 assert (gpu_data.get() == desired_result).all()
 
-    @mark_cuda_test
     def test_stride_preservation(self):
         A = np.random.rand(3, 3)
         AT = A.T
@@ -879,18 +979,15 @@ class TestGPUArray:
         print((AT_GPU.flags.f_contiguous, AT_GPU.flags.c_contiguous))
         assert np.allclose(AT_GPU.get(), AT)
 
-    @mark_cuda_test
     def test_vector_fill(self):
         a_gpu = gpuarray.GPUArray(100, dtype=gpuarray.vec.float3)
         a_gpu.fill(gpuarray.vec.make_float3(0.0, 0.0, 0.0))
         a = a_gpu.get()
         assert a.dtype == gpuarray.vec.float3
 
-    @mark_cuda_test
     def test_create_complex_zeros(self):
         gpuarray.zeros(3, np.complex64)
 
-    @mark_cuda_test
     def test_reshape(self):
         a = np.arange(128).reshape(8, 16).astype(np.float32)
         a_gpu = gpuarray.to_gpu(a)
@@ -923,7 +1020,6 @@ class TestGPUArray:
         a_gpu = a_gpu.reshape((4, 32))
         assert a_gpu.flags.c_contiguous
 
-    @mark_cuda_test
     def test_view(self):
         a = np.arange(128).reshape(8, 16).astype(np.float32)
         a_gpu = gpuarray.to_gpu(a)
@@ -940,7 +1036,6 @@ class TestGPUArray:
         view = a_gpu.view(np.int16)
         assert view.shape == (8, 32) and view.dtype == np.int16
 
-    @mark_cuda_test
     def test_squeeze(self):
         shape = (40, 2, 5, 100)
         a_cpu = np.random.random(size=shape)
@@ -957,7 +1052,8 @@ class TestGPUArray:
         assert a_gpu_squeezed_slice.flags.c_contiguous
 
         # Check that we get the original values out
-        assert np.all(a_gpu_slice.get().ravel() == a_gpu_squeezed_slice.get().ravel())
+        np.testing.assert_array_equal(a_gpu_slice.get().ravel(),
+                                      a_gpu_squeezed_slice.get().ravel())
 
         # Slice with length 1 on dimensions 2
         a_gpu_slice = a_gpu[:, :, 2:3, :]
@@ -970,9 +1066,9 @@ class TestGPUArray:
         assert not a_gpu_squeezed_slice.flags.c_contiguous
 
         # Check that we get the original values out
-        assert np.all(a_gpu_slice.get().ravel() == a_gpu_squeezed_slice.get().ravel())
+        np.testing.assert_array_equal(a_gpu_slice.get().ravel(),
+                                      a_gpu_squeezed_slice.get().ravel())
 
-    @mark_cuda_test
     def test_struct_reduce(self):
         preamble = """
         struct minmax_collector
@@ -1044,7 +1140,6 @@ class TestGPUArray:
         assert minmax["cur_min"] == np.min(a)
         assert minmax["cur_max"] == np.max(a)
 
-    @mark_cuda_test
     def test_reduce_out(self):
         from pycuda.curandom import rand as curand
 
@@ -1062,14 +1157,15 @@ class TestGPUArray:
 
         assert np.alltrue(a.max(axis=1) == max_gpu.get())
 
-    @mark_cuda_test
     def test_sum_allocator(self):
         # FIXME
         from pytest import skip
 
         skip("https://github.com/inducer/pycuda/issues/163")
-        # crashes with  terminate called after throwing an instance of 'pycuda::error'
-        # what():  explicit_context_dependent failed: invalid device context - no currently active context?
+        # crashes with  terminate called after throwing an instance
+        # of 'pycuda::error'
+        # what():  explicit_context_dependent failed: invalid device context -
+        # no currently active context?
 
         import pycuda.tools
 
@@ -1089,7 +1185,6 @@ class TestGPUArray:
         assert b.allocator == a.allocator
         assert c.allocator == pool.allocate
 
-    @mark_cuda_test
     def test_dot_allocator(self):
         # FIXME
         from pytest import skip
@@ -1121,7 +1216,6 @@ class TestGPUArray:
         assert dot_gpu_1.allocator == a_gpu.allocator
         assert dot_gpu_2.allocator == pool.allocate
 
-    @mark_cuda_test
     def test_view_and_strides(self):
         from pycuda.curandom import rand as curand
 
@@ -1134,7 +1228,6 @@ class TestGPUArray:
 
         assert np.array_equal(y.get(), X.get()[:3, :5])
 
-    @mark_cuda_test
     def test_scalar_comparisons(self):
         a = np.array([1.0, 0.25, 0.1, -0.1, 0.0])
         a_gpu = gpuarray.to_gpu(a)
@@ -1155,7 +1248,6 @@ class TestGPUArray:
         x = (a == 1).astype(a.dtype)
         assert (x == x_gpu.get()).all()
 
-    @mark_cuda_test
     def test_minimum_maximum_scalar(self):
         from pycuda.curandom import rand as curand
 
@@ -1171,17 +1263,14 @@ class TestGPUArray:
         assert la.norm(max_a0_gpu.get() - np.maximum(a, 0)) == 0
         assert la.norm(min_a0_gpu.get() - np.minimum(0, a)) == 0
 
-    @mark_cuda_test
     def test_transpose(self):
         from pycuda.curandom import rand as curand
 
         a_gpu = curand((10, 20, 30))
         a = a_gpu.get()
 
-        # assert np.allclose(a_gpu.transpose((1,2,0)).get(), a.transpose((1,2,0))) # not contiguous
         assert np.allclose(a_gpu.T.get(), a.T)
 
-    @mark_cuda_test
     def test_newaxis(self):
         from pycuda.curandom import rand as curand
 
@@ -1194,7 +1283,6 @@ class TestGPUArray:
         assert b_gpu.shape == b.shape
         assert b_gpu.strides == b.strides
 
-    @mark_cuda_test
     def test_copy(self):
         from pycuda.curandom import rand as curand
 
@@ -1233,7 +1321,6 @@ class TestGPUArray:
                 a_gpu.get()[start:stop:step, :, start:stop:step],
             )
 
-    @mark_cuda_test
     def test_get_set(self):
         import pycuda.gpuarray as gpuarray
 
@@ -1247,7 +1334,6 @@ class TestGPUArray:
         assert np.allclose(a_gpu.get(), a)
         assert np.allclose(a_gpu[1:3, 1:3, 1:3].get(), a[1:3, 1:3, 1:3])
 
-    @mark_cuda_test
     def test_zeros_like_etc(self):
         shape = (16, 16)
         a = np.random.randn(*shape).astype(np.float32)
@@ -1297,6 +1383,169 @@ class TestGPUArray:
                 assert new_z.dtype == np.complex64
                 assert new_z.shape == arr.shape
 
+    def test_logical_and_or(self):
+        rng = np.random.default_rng(seed=0)
+        for op in ["logical_and", "logical_or"]:
+            x_np = rng.random((10, 4))
+            y_np = rng.random((10, 4))
+            zeros_np = np.zeros((10, 4))
+            ones_np = np.ones((10, 4))
+
+            x_cu = gpuarray.to_gpu(x_np)
+            y_cu = gpuarray.to_gpu(y_np)
+            zeros_cu = gpuarray.zeros((10, 4), "float64")
+            ones_cu = gpuarray.ones((10, 4))
+
+            np.testing.assert_array_equal(
+                getattr(gpuarray, op)(x_cu, y_cu).get(),
+                getattr(np, op)(x_np, y_np))
+            np.testing.assert_array_equal(
+                getattr(gpuarray, op)(x_cu, ones_cu).get(),
+                getattr(np, op)(x_np, ones_np))
+            np.testing.assert_array_equal(
+                getattr(gpuarray, op)(x_cu, zeros_cu).get(),
+                getattr(np, op)(x_np, zeros_np))
+            np.testing.assert_array_equal(
+                getattr(gpuarray, op)(x_cu, 1.0).get(),
+                getattr(np, op)(x_np, ones_np))
+            np.testing.assert_array_equal(
+                getattr(gpuarray, op)(x_cu, 0.0).get(),
+                getattr(np, op)(x_np, 0.0))
+
+    def test_logical_not(self):
+        rng = np.random.default_rng(seed=0)
+        x_np = rng.random((10, 4))
+        x_cu = gpuarray.to_gpu(x_np)
+
+        np.testing.assert_array_equal(
+            gpuarray.logical_not(x_cu).get(),
+            np.logical_not(x_np))
+        np.testing.assert_array_equal(
+            gpuarray.logical_not(gpuarray.zeros(10, "float64")).get(),
+            np.logical_not(np.zeros(10)))
+        np.testing.assert_array_equal(
+            gpuarray.logical_not(gpuarray.ones(10)).get(),
+            np.logical_not(np.ones(10)))
+
+    def test_truth_value(self):
+        for i in range(5):
+            shape = (1,)*i
+            zeros = gpuarray.zeros(shape, dtype="float32")
+            ones = gpuarray.ones(shape, dtype="float32")
+            assert bool(ones)
+            assert not bool(zeros)
+
+    def test_setitem_scalar(self):
+        a = gpuarray.zeros(5, "float64") + 42
+        np.testing.assert_allclose(a.get(), 42)
+        a[...] = 1729
+        np.testing.assert_allclose(a.get(), 1729)
+
+    def test_default_zero(self):
+        # This test was added to make sure that
+        # gpurray.zeros is using np.float64 as the default dtype arg
+        a_gpu = gpuarray.zeros(10)
+        assert a_gpu.dtype == np.float64
+
+    @pytest.mark.parametrize("dtype,rtol", [(np.complex64, 1e-6),
+                                            (np.complex128, 1e-14)])
+    def test_log10(self, dtype, rtol):
+        from pycuda import cumath
+
+        rng = np.random.default_rng(seed=0)
+        x_np = rng.random((10, 4)) + dtype(1j)*rng.random((10, 4))
+        x_cu = gpuarray.to_gpu(x_np)
+        np.testing.assert_allclose(cumath.log10(x_cu).get(), np.log10(x_np),
+                                   rtol=rtol)
+
+    @pytest.mark.parametrize("ldtype", [np.int32, np.int64,
+                                        np.float32, np.float64,
+                                        np.complex64, np.complex128])
+    @pytest.mark.parametrize("rdtype", [np.int32, np.int64,
+                                        np.float32, np.float64,
+                                        np.complex64, np.complex128])
+    @pytest.mark.parametrize("op", [operator.add, operator.sub, operator.mul,
+                                    operator.truediv])
+    def test_binary_ops_with_unequal_dtypes(self, ldtype, rdtype, op):
+        # See https://github.com/inducer/pycuda/issues/372
+        if op == operator.truediv and {ldtype, rdtype} <= {np.int32, np.int64}:
+            pytest.xfail("Enable after"
+                         " gitlab.tiker.net/inducer/pycuda/-/merge_requests/66"
+                         "is merged.")
+
+        rng = np.random.default_rng(0)
+        lop_np = get_random_array(rng, (10, 4), ldtype)
+        rop_np = get_random_array(rng, (10, 4), rdtype)
+
+        expected_result = op(lop_np, rop_np)
+        result = op(gpuarray.to_gpu(lop_np), gpuarray.to_gpu(rop_np)).get()
+
+        assert result.dtype == expected_result.dtype
+        assert result.shape == expected_result.shape
+        np.testing.assert_allclose(expected_result, result,
+                                   rtol=5e-5)
+
+    def test_big_array_elementwise(self):
+        skip_if_not_enough_gpu_memory(4.5)
+
+        from pycuda.elementwise import ElementwiseKernel
+        n_items = 2**32
+
+        eltwise = ElementwiseKernel(
+            "unsigned char* d_arr",
+            "d_arr[i] = (unsigned char) (i & 0b11111111)", "mod_linspace"
+        )
+        d_arr = gpuarray.empty(n_items, np.uint8)
+        eltwise(d_arr)
+        result = d_arr.get()[()]
+        # Needs 8.6 GB memory on host - numpy cannot keep uint8 for mod() operation,
+        # and np.mod(np.arange()) is way too slow
+        reference = np.mod(np.arange(d_arr.size, dtype=np.int16), 256, dtype=np.int16)
+        reference -= result
+        assert np.max(reference) == 0
+
+    def test_big_array_reduction(self):
+        skip_if_not_enough_gpu_memory(4.5)
+
+        from pycuda.reduction import ReductionKernel
+        n_items = 2**32 + 11
+        reduction = ReductionKernel(
+            np.uint8,
+            neutral="0",
+            reduce_expr="(a+b) & 0b11111111",
+            map_expr="x[i]",
+            arguments="unsigned char* x"
+        )
+        d_arr = gpuarray.zeros(n_items, np.uint8)
+        d_arr.fill(1)  # elementwise!
+        result = reduction(d_arr).get()[()]
+
+        assert result == 11
+
+    def test_big_array_scan(self):
+        skip_if_not_enough_gpu_memory(4.5)
+        n_items = 2**32 + 12
+        from pycuda.scan import InclusiveScanKernel
+
+        cumsum = InclusiveScanKernel(np.uint8, "(a+b) & 0b11111111")
+        d_arr = gpuarray.zeros(n_items, np.uint8)
+        d_arr.fill(1)
+        result = cumsum(d_arr).get()[()]
+        # Needs 8.6 GB on host. numpy.allclose() is way too slow otherwise.
+        reference = np.tile(
+            np.roll(np.arange(256, dtype=np.int16), -1), n_items//256
+        )
+        reference -= result[:reference.size]
+        assert np.max(reference) == 0
+        assert np.allclose(result[2**32:], np.arange(1, 12+1))
+
+    def test_noncontig_transpose(self):
+        # https://github.com/inducer/pycuda/issues/385
+        d = gpuarray.zeros((1000, 15, 2048), "f")
+        d.transpose(axes=(1, 0, 2))  # works
+        d2 = d[:, 7:9, :]  # non C-contiguous
+        d2.transpose(axes=(1, 0, 2))  # crashes for recent versions
+
 
 if __name__ == "__main__":
     # make sure that import failures get reported, instead of skipping the tests.
